/*
   pikt.c -- PIKT script interpreter

   $Header: /var/cvs/PIKT/pikt/src/piktd/pikt.c,v 1.83 2007/01/02 09:13:42 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org

   --------------------------------------------------------------------------

   Usage:  pikt [-cGhTVz] [+M "<mailcmd>"] [+L "<lpcmd>"] [+C "<cmd>"] [+A|S <alert/script>]

           -c              syntax check alarm script file
           -G              run in debug mode
           -h              show program help
           -T              run in test mode
           -V              show version info
           -z              run in 'make check' mode
           +C              pikt command
           +M              mail command
           +L              print command
           +A|S            alert/script

 */

#include "../lib/common.h"
#include "piktd.h"

const char pikt_version[] = "PIKT pikt version " VERSION;

/* command-line options */
int ischeck = FALSE;
int isdebug = FALSE;
int isdoublecheck = FALSE;
int ishelp = FALSE;
int isshow_version = FALSE;
int istest = FALSE;
int ismakecheck = FALSE;

int bailout = FALSE;

int interactive = FALSE;

int isscript = FALSE;
int isalert = FALSE;

int istestsave;
int isdebugsave;

int scanpass;

char **argvcurr;
char **argvlast;

const char nullstring[] = { '\0' };

unsigned stroffset = 0;

int lineno = 0;

char *altcurr;

char *command = (char *) NULL;
char *ccp;

char *altlockfile;
char *execlockfile;
char *hstfile;

int dohistory = FALSE;

char *piktlog;
/* char syslog[] = "/var/log/syslog"; */

signed long arndx[MAXARYDIM+1];
signed long arndxnil[MAXARYDIM+1] = { NILARNDX, NILARNDX, NILARNDX, NILARNDX };
signed long arndxnul[MAXARYDIM+1] = {       0L,       0L,       0L,       0L };

const char *arytyp[] = { "ARYSTR", "ARYNUM" };	/* string exprinput prefixes
						   as a placeholder for an
						   array
						 */

const signed char err[]    = { ERROR, '\0' };
const signed char nil[]    = { NIL, '\0' };
const char errstr[] = "<<ERR>>";
const char nilstr[] = "<<NIL>>";

const signed char dquote[] = { DQTPROXY, '\0' };
const signed char squote[] = { SQTPROXY, '\0' };
const char space[]  = " ";

int exprtyp;

int exprres;
double exprval;
char *exprstr;

ino_t inputino = 0;
off_t inputsize = 0;
unsigned long inputposend = 0;
pid_t inputpid;

FILE *altin;

char *mailcmd, *lpcmd;

alrm *alarms = (alrm *) NULL;

alrm *alarmcurr = (alrm *) NULL;

val *valcurr;
ky *scvarcurr;
ky *keycurr;
rul *rulcurr = (rul *) NULL;
tok *tokcurr;
stm *stmcurr;

val **values[2][MAXVARS+1];

val *valp[MAXVAP];
val *valnow = (val *) NULL;
val *valprc = (val *) NULL;

int vapndx = 0;

unsigned long services = ALL_SERVICES;	/* all services enabled by default */

unsigned int fork_delay = 1;

void loadscripts(char *);
void beginalert(char *);
void endalert(char *);
void runscripts(char *);
int beginscript(void);
int runscript(void);
int endscript(void);

/* do program initializations; parse the command line; load the alarm scripts;
   run the scripts; send off alert messages, etc., as needed
 */
int
main (int argc, char *argv[])
{
	static char fnnam[] = "main";
	pid_t pid;
	static char pidstr[8];

	#if DEEPDEBUG
	char input[MAXLINE];
	char output[MAXLINE];

	while (TRUE) {
		gets(input);
		decompose(input, output);
		printf("%s\n", output);
	}
	#endif

	init(argc, argv, (void (*)(void))pikt_usage, pikt_version);
	if (bailout)
		exit(exstat);

	argvcurr = argv+1;	/* skip the program name */
	argvlast = argv+argc-1;

	cmdparse();		/* further parse the command line */
	if (bailout) {
		pikt_usage();
		exit(exstat);
	}

	if (! altcurr) {
		/* we haven't yet set piktlog at this point, so we log
		   this to syslog instead; we can safely set isautosyslog
		   to TRUE since we are exiting
		 */
		isautosyslog = TRUE;
		logexit(SYSLOG, LOG_ERR, ERRID_ALERT_NULL, __FILE__, __LINE__, fnnam, FALSE, "alert unknown (altcurr is NULL)");
	}
	if (isscript) {
		altlockfile = (char *) mymalloc((strlen(lokdir)+strlen(altcurr)+7) * sizeof(char), fnnam);
		sprintf(altlockfile, "%s/%s.lock", lokdir, altcurr);
	} else {
		altlockfile = (char *) mymalloc((strlen(lokdir)+strlen(altcurr)+11) * sizeof(char), fnnam);
		sprintf(altlockfile, "%s/%s.alt.lock", lokdir, altcurr);
	}
	execlockfile = (char *) mymalloc((strlen(lokdir)+strlen(altcurr)+12) * sizeof(char), fnnam);
	sprintf(execlockfile, "%s/%s.exec.lock", lokdir, altcurr);
	hstfile = (char *) mymalloc((strlen(hstdir)+strlen(altcurr)+6) * sizeof(char), fnnam);
	sprintf(hstfile, "%s/%s.hst", hstdir, altcurr);
	/* if logreturn() or logexit() are called before piktlog is set,
	   either should abort
	 */
	piktlog = (char *) mymalloc((strlen(logdir)+strlen(altcurr)+6) * sizeof(char), fnnam);
	sprintf(piktlog, "%s/%s.log", logdir, altcurr);

	/* these file checks come later than in the other PIKT binaries,
	   because isfile() can't log any stat() failure until after
	   piktlog is set
	 */
	if (    ! isfile(rmcmd, TRUE)
	     || ! isfile(tailcmd, TRUE)
	     || ! isfile(pscmd, TRUE)
	     || ! isfile(sumcmd, TRUE)
	     || ! isfile(cksumcmd, TRUE)
	   ) {
		/* error(s) logged in PIKTLOG */
		exit(exstat);
	}

	/* check for blocking lockfile */
	if (islock(altlockfile)) {
		if (interactive) {
			logexit(OUTPUT, LOG_ERR, ERRID_MULTIPLE_PIKT, __FILE__, __LINE__, fnnam, FALSE, "%s found!  Another %s pikt run may be in progress.  Try again later (else rm the %s lockfile).", altlockfile, altcurr, altlockfile);
		} else {
			logexit(PIKTLOG, LOG_ERR, ERRID_MULTIPLE_PIKT, __FILE__, __LINE__, fnnam, FALSE, "%s found!  Another %s pikt run may be in progress, run aborted.", altlockfile, altcurr);
		}
	}
	/* note: we can't now check if (procid("pikt", PROCID) == ERROR) as
	   with the other PIKT binaries, because of course different pikt's
	   can be running for different alerts; so we have to rely on the
	   lockfile mechanism alone for blocking multiple pikt's for the
	   same alert
	 */

	/* apply the lock, to prevent other concurrent pikt sessions
	   for this alert
	 */
	pid = getpid();
	sprintf(pidstr, "%u", (unsigned)pid);
	if (mklock(altlockfile, pidstr, interactive, TRUE) == ERROR)
		logexit(PIKTLOG, LOG_ERR, ERRID_FILE_CREATE_ERROR, __FILE__, __LINE__, fnnam, FALSE, "mklock() error for %s", altlockfile);

	loadscripts(altcurr);

	if (! bailout) {
		if (! ischeck && ! isdoublecheck) {
			beginalert(altcurr);
			runscripts(altcurr);
			endalert(altcurr);
		}
	}

	if (rmlocks(interactive, TRUE) == ERROR)
		logexit(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, FALSE, "rmlocks() error");

	exit(exstat);
}

/* output program usage to screen
 */
void
pikt_usage (void)
{
	printf("\nUsage:  pikt [-cGhTVz] [+M \"<mailcmd>\"] [+L \"<lpcmd>\"] [+C \"<cmd>\"] [+A|S <alert/script>]\n\n");
	printf("\t-c\t\tsyntax check alarm script file\n");
	printf("\t-G\t\trun in debug mode\n");
	printf("\t-h\t\tshow program help\n");
	printf("\t-T\t\trun in test mode\n");
	printf("\t-V\t\tshow version info\n");
	printf("\t-z\t\trun in 'make check' mode\n");
	printf("\t+C\t\tpikt command\n");
	printf("\t+M\t\tmail command\n");
	printf("\t+L\t\tprint command\n");
	printf("\t+A|S\t\talert/script\n\n");
	return;
}

/* load and parse the alarm script file
 */
void
loadscripts (char *alert)
{
	static char fnnam[] = "loadscripts";
	char *altfile;

	if (! command) {
		if (isdoublecheck) {
			if (isscript) {
				altfile = (char *) mymalloc((strlen(stgdir)+strlen(alert)+2) * sizeof(char), fnnam);
				sprintf(altfile, "%s/%s", stgdir, alert);
			} else {
				altfile = (char *) mymalloc((strlen(stgdir)+strlen(alert)+6) * sizeof(char), fnnam);
				sprintf(altfile, "%s/%s.alt", stgdir, alert);
			}
		} else {
			if (isscript) {
				altfile = (char *) mymalloc((strlen(prgdir)+strlen(alert)+2) * sizeof(char), fnnam);
				sprintf(altfile, "%s/%s", prgdir, alert);
			} else {
				altfile = (char *) mymalloc((strlen(altdir)+strlen(alert)+6) * sizeof(char), fnnam);
				sprintf(altfile, "%s/%s.alt", altdir, alert);
			}
		}
		if ((altin = fopen(altfile, "r")) == NULL) {
			if (isscript) {
				free(altfile);
				/* now try with '.pkt' appended */
				altfile = (char *) mymalloc((strlen(prgdir)+strlen(alert)+6) * sizeof(char), fnnam);
				sprintf(altfile, "%s/%s.pkt", prgdir, alert);
				if ((altin = fopen(altfile, "r")) == NULL) {
					logexit(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", altfile);
				}
			} else {
				logexit(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", altfile);
			}
		}
	} else {
		altfile = mystrdup(nullstring, fnnam);
	}
	lineno = 1;
	if (isalert)
		alarmcurr = (alrm *) NULL;
	scanpass = 1;
	if (command)
		ccp = command;
	altparse();
	if (bailout)
		return;
	if (! command) {
		rewind(altin);
		if (! alarmcurr)
			logexit(PIKTLOG, LOG_ERR, ERRID_ALARM_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "unknown alarm/script name for %s", altfile);
		if (isalert)
			alarmcurr = (alrm *) NULL;
	}
	scanpass = 2;
	if (command)
		ccp = command;
	altparse();
	if (! command) {
		if (fclose(altin))	/* returns 0 on success */
			logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", altfile);
		free(altfile);
	}
	return;
}

/* open the history file, if necessary; do some alert initializations
 */
void
beginalert (char *alert)
{
	static char fnnam[] = "beginalert";
	struct stat fbuf;
	register int i;

	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "%s run begin", isscript ? "script" : "alert");
	if (dohistory) {
		openhistory(hstfile);
	} else {	/* no current history, so remove any old history file */
		if (stat(hstfile, &fbuf) == 0) {
			if (remove(hstfile))
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "remove() error for %s", hstfile);
		}
	}
	keyseplen = strlen(KEYSEP);
	mailhdrmsgcurr = mailhdrmsg = (msgq *) NULL;
	printhdrmsgcurr = printhdrmsg = (msgq *) NULL;
	for (i=0; i<8; i++) {
		mailalmmsgcurr[i] = mailalmmsg[i] = (msgq *) NULL;
		printalmmsgcurr[i] = printalmmsg[i] = (msgq *) NULL;
	}
	newmailalert = TRUE;
	newprintalert = TRUE;
	return;
}

/* send alert email messages, if necessary; send printer output, if necessary;
   close the history file, if necessary
 */
void
endalert (char *alert)
{
	static char fnnam[] = "endalert";
	msgq *mp;
	FILE *altfile;
	register int i;

	if (mailhdrmsg) {
		if (mailcmd) {
			if ((altfile = popen(mailcmd, "w")) == NULL)
				logexit(PIKTLOG, LOG_ERR, ERRID_PROC_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, popen() error for %s", ALMNAM, mailcmd);
			foreach (mp, mailhdrmsg)
				fprintf(altfile, "%s", mp->line);
			for (i=0; i<8; i++) {
				foreach (mp, mailalmmsg[i])
					fprintf(altfile, "%s", mp->line);
			}
			pclose(altfile);
			freemsg(&mailhdrmsg);
			for (i=0; i<8; i++) {
				freemsg(&mailalmmsg[i]);
			}
		} else {
			logreturn(PIKTLOG, LOG_ERR, ERRID_MSG_CMD_NULL, __FILE__, __LINE__, fnnam, FALSE, "for alert %s, messages queued, but missing mailcmd", altcurr);
		}
	}
	if (printhdrmsg) {
		if (lpcmd) {
			if ((altfile = popen(lpcmd, "w")) == NULL)
				logexit(PIKTLOG, LOG_ERR, ERRID_PROC_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, popen() error for %s", ALMNAM, lpcmd);
			foreach (mp, printhdrmsg)
				fprintf(altfile, "%s", mp->line);
			for (i=0; i<8; i++) {
				foreach (mp, printalmmsg[i])
					fprintf(altfile, "%s", mp->line);
			}
			pclose(altfile);
			freemsg(&printhdrmsg);
			for (i=0; i<8; i++) {
				freemsg(&printalmmsg[i]);
			}
		} else {
			logreturn(PIKTLOG, LOG_ERR, ERRID_MSG_CMD_NULL, __FILE__, __LINE__, fnnam, FALSE, "for alert %s, messages queued, but missing lpcmd", altcurr);
		}
	}
	if (dohistory)
		closehistory(hstfile);
	free(hstfile);
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "%s run end", isscript ? "script" : "alert");
	return;
}

/* for each alarm script:  initialize; open script input, if called for;
   begin writing the alarm history, if called for; run the script; close
   script input, if called for; finish writing the alarm history, if called
   for; tidy up
 */
void
runscripts (char *alert)
{
	static char fnnam[] = "runscripts";

	istestsave = istest;
	isdebugsave = isdebug;
	foreach (alarmcurr, alarms) {
		istest = istestsave;		/* reset */
		isdebug = isdebugsave;		/* reset */
		switch (STATUS) {
			case ERROR:
				logreturn(PIKTLOG, LOG_ERR, ERRID_ALARM_ERROR_STATUS, __FILE__, __LINE__, fnnam, FALSE, "%s alarm in error status", ALMNAM);
				continue;
			case INACTIVE:
				if (isverboselog)
					logreturn(PIKTLOG, LOG_INFO, ERRID_ALARM_INACTIVE_STATUS, __FILE__, __LINE__, fnnam, FALSE, "%s alarm in inactive status", ALMNAM);
				continue;
			case SUSPENDED:
				logreturn(PIKTLOG, LOG_NOTICE, ERRID_ALARM_SUSPENDED_STATUS, __FILE__, __LINE__, fnnam, FALSE, "%s alarm in suspended status", ALMNAM);
				continue;
			case TESTING:
				/* if (isverboselog) */
					logreturn(PIKTLOG, LOG_INFO, ERRID_ALARM_TESTING_STATUS, __FILE__, __LINE__, fnnam, FALSE, "%s alarm in testing status", ALMNAM);
				istest = TRUE;
				break;
			case DEBUGGING:
				/* if (isverboselog) */
					logreturn(PIKTLOG, LOG_INFO, ERRID_ALARM_DEBUGGING_STATUS, __FILE__, __LINE__, fnnam, FALSE, "%s alarm in debugging status", ALMNAM);
				isdebug = TRUE;
				break;
			case ACTIVE:
				break;
			case NONE:
			default:
				logreturn(PIKTLOG, LOG_WARNING, ERRID_ALARM_UNKNOWN_STATUS, __FILE__, __LINE__, fnnam, FALSE, "%s alarm in unknown status", ALMNAM);
				continue;
		}
		if (beginscript() == ERROR)
			continue;
		switch (runscript()) {
			case ERROR:
				logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "aborted %s", ALMNAM);
				break;
			case QUIT:
				logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "quit %s", ALMNAM);
				break;
			case 0:
				logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "finished %s", ALMNAM);
				break;
		}
		endscript();
	}
	return;
}

/* run the script begin section; and if no errors (and no quit), run the
   script rules section; and if no errors (and no quit), run the script end
   section
 */
int
runscript (void)
{
	FILE *infile = (FILE *) NULL;
	FILE *intmpfile = (FILE *) NULL;
	int res;

	if ((res = dobegin()) == 0) {
		if ((res = openinput(&infile, &inputpid, INPUTTYP, INPUTSRC, &intmpfile, INPUTTMPFIL)) == ERROR)
			return res;
		if ((res = beginhistory()) == ERROR) {
			closeinput(infile, inputpid, intmpfile);
			return res;
		}
		if ((res = dorules(INPUTTYP, infile, POSTYP, intmpfile)) == 0) {
			res = doend();
		}
		closeinput(infile, inputpid, intmpfile);
		endhistory();
	}
	return res;
}

/* remove any potentially blocking exec lock file; do some initializations
 */
int
beginscript (void)
{
	static char fnnam[] = "beginscript";

	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "ALARM %s", ALMNAM);
	newmailalarm = TRUE;
	newprintalarm = TRUE;
	/* remove any potentially blocking exec lock file
	if (islock(execlockfile)) {
		if (rmlock(execlockfile, interactive, TRUE) == ERROR)
			logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, FALSE, "rmlock() error for %s", execlockfile);
	}
	 */
	if (! isscript)
		initvals(&valnow);
	/* else initvals(&valnow) already done once and for all in cmd.y
	   to prep for $ARGV[] and #ARGC
	 */
	initvals(&valprc);
	#if DEEPDEBUG
	if ((HSTTYP & (FIXEDVAL|FREEVAL)) && ! HSTERR) {
		if (isdebug)
			dumpvals(VALHST);
	}
	#endif
	return 0;
}

/* close any open files; garbage collect all unneeded script values; in
   input file rewrite is indicated, make a backup copy of the input file,
   then substitute the input temporary file for input file
 */
int
endscript (void)
{
	static char fnnam[] = "endscript";
	register int i;
	vl *vlp;
	char *cmd;
	char *res;
	int resnum;
	int bakfilerr;
	#ifdef AIX
	char *scs;              /* source checksum string */
	char *tcs;              /* target checksum string */
	#endif

	/* close any open files */
	for (i=1; i<=VRNDXHI; i++) {
		vlp = lookupvl(valnow, i, arndxnul, (char *)NULL, ARYTYP(i));
		if (((DATTYP(i) == FILEHANDLE) || (DATTYP(i) == PROCHANDLE)) && vlp && vlp->fil) {
			if (fclose(vlp->fil))
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fclose() error for file handle %s", ALMNAM, VARNAM(i));
		}
	}
	/* we no longer need any of the vals--whether valnows, valprcs, or
	   valhst, so let's free the memory allocated to them; 
	 */
	freevals(&valnow, FIXEDVAL|FREEVAL|BEGINVAL|ENDVAL);
	freevals(&valprc, FIXEDVAL|FREEVAL|BEGINVAL|ENDVAL);
	freevals(&(VALHST), FIXEDVAL|FREEVAL|BEGINVAL|ENDVAL);
	/* if an input file rewrite is indicated, copy INPUTSRC (the input
	   file) to INPUTBAKFIL (the backup file), and replace INPUTSRC
	   with INPUTTMPFIL (the rewritten input file)
	*/
	bakfilerr = FALSE;
	if (REWRITEINFIL) {
		if (STATUS == ACTIVE) {
			cmd = (char *) mymalloc((size_t)(strlen(cpcmd)+strlen(INPUTSRC)+strlen(INPUTBAKFIL)+6) * sizeof(char), fnnam);
			sprintf(cmd, "%s -p %s %s", cpcmd, INPUTSRC, INPUTBAKFIL);
			/* cp -p <inputfile> <bakfile> */
			res = sys(cmd,SYSTEM);
			#ifdef AIX
			/* to compensate for the broken sys() function;
			   just compare MD5 checksums and don't bother to
			   check time/date stamps
			 */
			scs = checksumfil(5, INPUTSRC);
			tcs = checksumfil(5, INPUTBAKFIL);
			if (csstrcmp(scs, tcs)) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_COPY_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, csstrcmp() returned nonzero, failed copy command: %s", ALMNAM, cmd);
				bakfilerr = TRUE;
			}
			free(scs);
			free(tcs);
			#else
			if (! strcmp(res, err)) {
				bakfilerr = TRUE;
			} else {
				errno = 0;
				resnum = atoi(res);
				if (errno) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_COPY_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, failed copy command: %s", ALMNAM, cmd);
					bakfilerr = TRUE;
				} else {
					if (resnum) {
						logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_COPY_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, failed copy command: %s", ALMNAM, cmd);
						bakfilerr = TRUE;
					}
				}
			}
			#endif
			free(cmd);
			free(res);
			if (! bakfilerr) {
				/* rename <intmpfile> -> <inputfile> */
				logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "renaming %s as %s", INPUTTMPFIL, INPUTSRC);
				if ((resnum = rename(INPUTTMPFIL, INPUTSRC))) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_RENAME_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, rename() failure of %s to %s", ALMNAM, INPUTTMPFIL, INPUTSRC);
				return ERROR;
				}
			} else {
				if (unlink(INPUTTMPFIL)) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, unlink() failure for %s", ALMNAM, INPUTTMPFIL);
					return ERROR;
				}
				return ERROR;
			}
		}
	} else {
		if (INPUTTMPFIL && filestat(INPUTTMPFIL, EXIST)) {
			if (unlink(INPUTTMPFIL)) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, unlink() failure for %s", ALMNAM, INPUTTMPFIL);
				return ERROR;
			}
		}
	}
	return 0;
}
